/*
 * [FastCat.java]
 *
 * Summary: Stripped down, fast replacement for StringBuilder and StringWriter to build concatenated Strings.
 *
 * Copyright: (c) 2009-2013 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2009-09-29 initial release
 *  1.1 2009-10-09 return this from append
 *  1.2 2010-01-25 fix bug in rarely used clear method.
 *  1.3 2010-02-13 convert from JDK 1.6 to JDK 1.5
 *  1.4 2011-06-08 add room, size, used methods
 *  1.5 2011-11-13 append methods now ignore null parameters
 *  1.6 2011-11-19 checkEstimate method to get advance warning an estimate is too tight or too generous.
 *  1.7 2012-11-19 add the drop method to drop last appended String. e.g. to trim last comma.
 */
package com.mindprod.fastcat;

import java.io.File;
import java.util.Arrays;

import static java.lang.System.err;
import static java.lang.System.out;

/**
 * Stripped down, fast replacement for StringBuilder and StringWriter to build concatenated Strings.
 * <p/>
 * For concatenating individual characters, StringBuilder will probably be faster.
 * It has an overhead of 4 bytes per piece appended, which would be 50% if the Strings were only 2-bytes long.
 * This algorithm works best with fairly long pieces. The main advantage is it is much easier to get a precise
 * estimate of how much space you will need.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.7 2012-11-19 add the drop method to drop last appended String. e.g. to trim last comma.
 * @since 2009-09-29
 */
public class FastCat
    {
    // ------------------------------ CONSTANTS ------------------------------

    /**
     * default estimated number of pieces
     */
    private static final int DEFAULT_SIZE = 1000;

    /**
     * how much extra we give to each estimate of space required to ensure small files still get some growth.
     */
    private static final int EXTRA = 50;

    private static final int FIRST_COPYRIGHT_YEAR = 2009;

    /**
     * If the estimate is 30% bigger than required, we consider it overly generous
     */
    private static final int GENEROUS_PERCENT = 130;

    /**
     * if we get an overflow, we suggest doubling the size estimate.
     */
    private static final int RECOVER_PERCENT = 200;

    /**
     * If the estimate is too tight or too generous, we suggest one 20% bigger than required.
     */
    private static final int SUGGESTED_PERCENT = 120;

    /**
     * If as estimate is not at least 10% bigger than required, we consider it too tight.
     */
    private static final int TIGHT_PERCENT = 110;

    /**
     * undisplayed copyright notice
     */
    @SuppressWarnings( { "UnusedDeclaration" } )
    private static final String EMBEDDED_COPYRIGHT =
            "Copyright: (c) 2009-2013 Roedy Green, Canadian Mind Products, http://mindprod.com";

    /**
     * when package was released.
     *
     * @noinspection UnusedDeclaration
     */
    private static final String RELEASE_DATE = "2012-11-19";

    /**
     * version of package.
     *
     * @noinspection UnusedDeclaration
     */
    private static final String VERSION_STRING = "1.7";

    // ------------------------------ FIELDS ------------------------------

    /**
     * private collection of pieces we will later glue together once we know the size of the final String.
     * For RAM efficiency, we use an array, not an ArrayList
     */
    private final String[] pieces;

    /**
     * how many pieces we have collected pending so far
     */
    private int count = 0;

    /**
     * total length of all the  pieces we have collected pending so far
     */
    private int length = 0;

    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * no-arg constructor, defaults to 20 pieces as the estimate.
     */
    public FastCat()
        {
        this( DEFAULT_SIZE );
        }

    /**
     * constructor
     *
     * @param estNumberOfPieces estimated number of chunks you will concatenate.
     *                          If the estimate is low, you will get ArrayIndexOutOfBoundsExceptions.
     */
    public FastCat( int estNumberOfPieces )
        {
        pieces = new String[ estNumberOfPieces ];
        }

    /**
     * append String
     *
     * @param s String to append.
     *          If s is null, nothing happens.
     *          No NullPointerException.
     *          Nothing added to the buffer.
     *
     * @return this
     */
    public FastCat append( String s )
        {
        if ( s == null )
            {
            return this;
            }
        if ( count >= pieces.length )
            {
            overflow();
            }
        length += s.length();
        pieces[ count++ ] = s;
        return this;
        }

    /**
     * append arbitrary number of strings
     *
     * @param ss comma-separated list of Strings to append
     *
     * @return this
     */
    public FastCat append( String... ss )
        {
        if ( ss == null )
            {
            return this;
            }
        for ( String s : ss )
            {
            if ( count >= pieces.length )
                {
                overflow();
                }
            length += s.length();
            pieces[ count++ ] = s;
            }
        return this;
        }

    /**
     * append int
     *
     * @param i int to append.
     *
     * @return this
     */
    public FastCat append( int i )
        {
        return append( Integer.toString( i ) );
        }

    /**
     * append char
     *
     * @param c char to append.
     *          If you use this method extensively, you will probably get better performance from StringBuilder.
     *
     * @return this
     */
    public FastCat append( char c )
        {
        if ( count >= pieces.length )
            {
            overflow();
            }
        final String s = String.valueOf( c );
        length += s.length();
        pieces[ count++ ] = s;
        return this;
        }

    /**
     * append Object
     *
     * @param o Object to append.  toString is called to acquire a String to concatenate.
     *
     * @return this
     */
    public FastCat append( Object o )
        {
        if ( o == null )
            {
            return this;
            }
        if ( count >= pieces.length )
            {
            overflow();
            }
        final String s = o.toString();
        length += s.length();
        pieces[ count++ ] = s;
        return this;
        }

    /**
     * append arbitrary number of Objects
     *
     * @param oo comma-separated list of Objects to to append. toString is called to acquire a String to concatenate.
     *
     * @return this
     */
    public FastCat append( Object... oo )
        {
        if ( oo == null )
            {
            return this;
            }
        for ( Object o : oo )
            {
            if ( count >= pieces.length )
                {
                overflow();
                }
            final String s = o.toString();
            length += s.length();
            pieces[ count++ ] = s;
            }
        return this;
        }

    /**
     * Optionally check that FastCat size estimate was reasonable, not too tight, not too generous.
     * Call just prior to calling toString.  If FastCat overflows before you get around to calling
     * checkEstimate, you will get an ArrayIndexOutOfBoundsException. In that case, this method will to be involved.
     *
     * @param whereTextCameFrom where generated text came from, some text to use in the warning message printed on
     *                          System.err.
     */
    public void checkEstimate( final String whereTextCameFrom )
        {
        final int room = room();
        final int used = used();
        final int size = size();
        final int tightLimit = used * TIGHT_PERCENT / 100 + EXTRA;
        final int generousLimit = used * GENEROUS_PERCENT / 100 + EXTRA;
        final int suggestedSize = used * SUGGESTED_PERCENT / 100 + EXTRA;
        // don't complain if has it bang on or off by up to 5.
        // Presume aiming for bang on rather than leaving room for growth.
        if ( room > 5 )
            {
            if ( size < tightLimit )
                {
                err.println( "\n>>>> WARNING >>>> FastCat estimate tight for " + whereTextCameFrom + "\n" +
                             "                  used: " + used + " + room: " + room + " = size: " + size + " suggest " +
                             "size: " + suggestedSize );
                }
            // For default size, we don't worry if overly generous.
            else if ( room > generousLimit && size != DEFAULT_SIZE )
                {
                err.println( "\n>>>> WARNING >>>> FastCat estimate generous for " + whereTextCameFrom + "\n" +
                             "                  used: " + used + " + room: " + room + " = size: " + size + " suggest " +
                             "size: " + suggestedSize );
                }
            }
        }

    /**
     * empty the concatenated String being created
     */
    public void clear()
        {
        // clear out strings so they can be garbage collected. All would work without fill, but it would waste RAM.
        // works ok even if count is 0.
        // are we are doing is setting each of the pieces to a null, thus freeing the associated string for gc
        Arrays.fill( pieces, 0, count, null );
        count = 0;
        length = 0;
        }

    /**
     * drops the last appended string, not the last char. If none, does nothing.
     * Useful for trimming last comma when generating a comma-separated list.
     * May be called repeatedly to peel off previously appended Strings.
     * It is a terse way of getting the same effect as:
     * if ( sb.used() != 0 )
     * {
     * sb.append( ", " );
     * }
     * sb.append( nextElt );
     *
     * @return this
     */
    public FastCat drop()
        {
        count--;
        if ( count < 0 )
            {
            count = 0;
            }
        else
            {
            length -= pieces[ count ].length();
            pieces[ count ] = null;
            }
        return this;
        }

    /**
     * current buffer length.
     *
     * @return current total of count of chars appended
     * @see #used()
     */
    public int length()
        {
        return length;
        }

    /**
     * how many unused slots are available for pieces?
     * Not same as total number of chars that could append.
     *
     * @return count of unused slots.
     */
    public int room()
        {
        return pieces.length - count;
        }

    /**
     * how many slots (used or unused) are available for pieces?
     * Not same as total number of chars that could append.
     *
     * @return count of unused/unused slots.
     * @see #length()
     * @see #used()
     */
    public int size()
        {
        return pieces.length;
        }

    /**
     * Get the concatenation of all the strings appended so far
     */
    public String toString()
        {
        int offset = 0;
        final char[] buffer = new char[ length ];
        for ( int i = 0; i < count; i++ )
            {
            final String piece = pieces[ i ];
            // copy chars from String into our accumulating char[] buffer with System.arraycopy.
            piece.getChars( 0, piece.length(), buffer, offset );
            offset += piece.length();
            }
        return new String( buffer );  // Would like some way to just hand buffer over to String to avoid copy.
        }

    /**
     * how many used slots, slots containing pieces?
     * Not same as total number of chars appended.
     *
     * @return count of unused/unused slots.
     * @see #length()
     * @see #size()
     */
    public int used()
        {
        return count;
        }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * We overflowed the array.  Give up.  Get user to improve estimate.  We could recover automatically,
     * but that would defeat the intent of fast code.
     */
    private void overflow()
        {
        // Sun JVM lets us figure out where the call came from.
        StackTraceElement e = new Throwable().getStackTrace()[ 2 ];
        // we know where in the code, but not which file was being processed, or any other user-supplied indication.
        final String whereTextCameFrom = "class.method: "
                                         + e.getClassName()
                                         + "."
                                         + e.getMethodName()
                                         + " line:"
                                         + e.getLineNumber();

        final int room = room();
        final int used = used();
        final int size = size();
        final int recoverSize = used * RECOVER_PERCENT / 100 + EXTRA;
        // don't complain about small FastCats that are bang on or nearly bang on.

        err.println( "\n>>>> ERROR >>>> FastCat estimate catastrophically low for " + whereTextCameFrom + "\n" +
                     "                  used: " + used + " + room: " + room + " = size: " + size + " suggest size: "
                     + recoverSize + "\n" +
                     "                  see below for additional information on where this happened.\n" );

        throw new ArrayIndexOutOfBoundsException( "FastCat Overflow" );
        }

    // --------------------------- main() method ---------------------------

    /**
     * test harness
     *
     * @param args not used.
     */
    public static void main( String[] args )
        {
        final FastCat sb = new FastCat( 7 );
        sb.append( "Hello" );
        sb.append( " " );
        sb.append( "World. " );
        sb.append( new File( "temp.txt" ) );
        sb.append( " ", "abc", "def" );
        out.println( sb.toString() );
        // prints Hello World. temp.txt abcdef
        }
    }
